#!/usr/bin/env python
"""
Setup a virtual machine from a config file.

Note: Most of the operations requires root access and this script uses ``sudo``
to get them.

"""
import argparse
import os
import subprocess
import sys
import tempfile
import time


from bzrlib import (
    config,
    osutils,
    transport,
    urlutils,
    )

# Work around a bug in bzrlib.config forbidding some constructs in templates.
# Namely, spaces are invalid as an identifier and therefore should not match
# below.
config._option_ref_re = config.lazy_regex.lazy_compile('({[^ {}\n]+})')


class VmMatcher(config.NameMatcher):

    def match(self, section):
        if section.id is None:
            # The no name section contains default values
            return True
        return super(VmMatcher, self).match(section)

    def get_sections(self):
        matching_sections = super(VmMatcher, self).get_sections()
        return reversed(list(matching_sections))


class VmStore(config.LockableIniFileStore):
    """A config store for options specific to a directory."""

    def __init__(self, directory, file_name, possible_transports=None):
        t = transport.get_transport_from_path(
            directory, possible_transports=possible_transports)
        super(VmStore, self).__init__(t, file_name)
        self.id = 'vm'


class VmStack(config.Stack):
    """Per-directory options."""

    def __init__(self, name):
        """Make a new stack for a given vm.

        :param name: The name of a virtual machine."""
        dstore = VmStore(os.environ['HOME'], 'vms.conf')
        # FIXME: Only available in bzr-2.6b3 :-/ -- vila 2012-01-31
        # dstore = self.get_shared_store()
        super(VmStack, self).__init__(
            [VmMatcher(dstore, name).get_sections,],
            dstore, mutable_section_id=name)


def register(option):
    config.option_registry.register(option)



# The base directories where vms are stored
register(config.Option(
        'vm.images_dir', default='/var/lib/libvirt/images',
        help="Where vm disk images are stored."
        ))
register(config.Option(
        'vm.qemu_etc_dir',
        default='/etc/libvirt/qemu',
        help="Where libvirt (qemu) stores the vms config files."
        ))

# Isos and images download handling
register(config.Option(
        'vm.iso_url',
        default='http://cdimage.ubuntu.com/daily-live/current/' ,
        help="Where an iso can be downloaded from."
        ))
register(config.Option(
        'vm.iso_name',
        default='{vm.release}-desktop-{vm.cpu_model}.iso',
        help="The name of the iso."
        ))
register(config.Option(
        'vm.cloud_image_url',
        default='http://cloud-images.ubuntu.com/{vm.release}/current/',
        help="Where a cloud image can be downloaded from."
        ))
register(config.Option(
        'vm.cloud_image_name',
        default='{vm.release}-server-cloudimg-{vm.cpu_model}-disk1.img',
        help="The name of the cloud image."
        ))
register(config.Option(
        'vm.download_cache',
        default='{vm.images_dir}',
        help="Where downloads end up."
        ))

# The options required to create a vm
register(config.Option(
        'vm.name', default=None, invalid='error',
        help="The vm name, used as a prefix for related files."
        ))
register(config.Option(
        'vm.ram_size', default='1024',
        help="The ram size in megabytes."
        ))
register(config.Option(
        'vm.disk_size', default='8G',
        help="The ram size in megabytes."
        ))
register(config.Option(
        'vm.cpus', default='1',
        help="The number of cpus."
        ))
register(config.Option(
        'vm.cpu_model', default=None, invalid='error',
        help="The number of cpus."))
register(config.Option(
        'vm.network', default='network=default', invalid='error',
        help="""The --network parameter for virt-install.

This can be specialized for each machine but the default should work in most
setups. Watch for your DHCP server exhausting its address space if you create a
lot of vms with random MAC addresses.
"""))
register(config.Option(
        'vm.password', default='ubuntu', invalid='error',
        help="The ubuntu user password."
        ))
register(config.Option(
        'vm.release', default=None, invalid='error',
        help="The ubuntu release name."
        ))
register(config.Option(
        'vm.meta_data', default='''\
instance-id: {vm.name}
local-hostname: {vm.name}
''',
        invalid='error',
        help="The meta data for cloud-init to put in the seed."
        ))
register(config.Option(
        'vm.user_data', default='''\
#cloud-config
locale: en_US.UTF-8

password: {vm.password}
chpasswd: { expire: False }
ssh_pwauth: True

ssh_authorized_keys:
 - {vm.ssh_pub_key}

{vm.apt_proxy}

apt_sources: 
 - source: deb http://archive.ubuntu.com/ubuntu {vm.release} restricted multiverse
 - source: deb-src http://archive.ubuntu.com/ubuntu {vm.release} restricted multiverse

apt_update: true
apt_upgrade: true
byobu_by_default: system

manage_etc_hosts: true

packages:
 - bzr
 - ubuntu-desktop
 - openssh-server
 - avahi-daemon

''',
        invalid='error',
        help="The user data for cloud-init to put in the seed."
        ))

# Some bits that may expanded in user-data but are optional

register(config.Option(
        'vm.apt_proxy', default='', invalid='error',
        help='''A local proxy for apt to avoid repeated .deb downloads.

It should start with 'apt_proxy: ' if set in any fashion.

Example:

  vm.apt_proxy = apt_proxy: http://192.168.0.42:8000

'''
        ))


class SetupVmError(Exception):

    msg = 'setup_vm Generic Error: %r'

    def __init__(self, msg=None, **kwds):
        if msg is not None:
            self.msg = msg
        for key, value in kwds.items():
            setattr(self, key, value)

    def __str__(self):
        return self.msg.format((), **self.__dict__)

    __repr__ = __str__


class CommandError(SetupVmError):

    msg = '''
  command: {joined_cmd}
  retcode: {retcode}
  output: {out}
  error: {err}
'''

    def __init__(self, cmd, retcode, out, err):
        super(CommandError, self).__init__(joined_cmd=' '.join(cmd),
                                           retcode=retcode, err=err, out=out)

class ConfigValueError(SetupVmError):

    msg = 'Bad value "{value}" for option "{name}".'

    def __init__(self, name, value):
        super(ConfigValueError, self).__init__(name=name, value=value)


def run_subprocess(args):
    proc = subprocess.Popen(args,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            stdin=subprocess.PIPE)
    out, err = proc.communicate()
    if proc.returncode:
        raise CommandError(args, proc.returncode, out, err)
    return proc.returncode, out, err

def pipe_subprocess(args):
    proc = subprocess.Popen(args,
                            stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    return proc


def vm_states(source=None):
    """A dict of states for vms indexed by name.

    :param source: A list of lines as produced by virsh list --all without
        decorations (header/footer).
    """
    if source is None:
        retcode, out, err = run_subprocess(['virsh', 'list', '--all'])
        # Get rid of header/footer
        source = out.splitlines()[2:-1]
    states = {}
    for line in source:
        caret_or_id, name, state = line.split(None, 2)
        states[name] = state
    return states


class VM(object):

    def __init__(self, conf):
        self.conf = conf

    def run_local(self, args):
        proc = subprocess.Popen(args,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE,
                                stdin=subprocess.PIPE)
        out, err = proc.communicate()
        if proc.returncode:
                raise CommandError(args, proc.returncode, out, err)
        return proc.returncode, out, err



class Kvm(VM):

    def __init__(self, conf):
        super(Kvm, self).__init__(conf)
        # Seed files
        self._meta_data_file = None
        self._user_data_file = None
        self._seed_path = None
        self._disk_image_path = None
        self._console_path = None

    def _download_in_cache(self, source_url, name, force=False):
        """Download ``name`` from ``source_url`` in ``vm.download_cache``.

        :param source_url: The url where the file to download is located

        :param name: The name of the file to download (also used as the name
            for the downloaded file).

        :param force: Remove the file from the cache if present.

        :return: False if the file is in the download cache, True if a download
            occurred.
        """
        source = urlutils.join(source_url, name)
        download_dir = self.conf.get('vm.download_cache')
        if not os.path.exists(download_dir):
            raise ConfigValueError('vm.download_cache', download_dir)
        target = os.path.join(download_dir, name)
        # FIXME: By default the download dir may be under root control, but if
        # a user chose to use a different one under his own control, it would
        # be nice to not require sudo usage. -- vila 2013-02-06
        if force:
            run_subprocess(['sudo', 'rm', '-f', target])
        if not os.path.exists(target):
            # FIXME: We do ask for a progress bar but it's not displayed
            # (run_subprocess capture both stdout and stderr) ! At least while
            # used interactively, it should. -- vila 2013-02-06
            run_subprocess(['sudo', 'wget', '--progress=dot:mega','-O',
                            target, source])
            return True
        else:
            return False

    def download_iso(self, force=False):
        """Download the iso to install the vm.

        :return: False if the iso is in the download cache, True if a download
            occurred.
        """
        return self._download_in_cache(self.conf.get('vm.iso_url'),
                                       self.conf.get('vm.iso_name'),
                                       force=force)

    def download_cloud_image(self, force=False):
        """Download the cloud image to install the vm.

        :return: False if the image is in the download cache, True if a
            download occurred.
        """
        return self._download_in_cache(self.conf.get('vm.cloud_image_url'),
                                       self.conf.get('vm.cloud_image_name'),
                                       force=force)

    def _create_tempfile_with_content(self, content):
        temp = tempfile.NamedTemporaryFile()
        # We don't close the file here. This ensures it will be deleted when
        # self goes out of scope freeing us to handle the cleanup.
        temp.file.write(content)
        temp.file.flush()
        return temp

    def create_meta_data(self):
        self._meta_data_file = self._create_tempfile_with_content(
            self.conf.get('vm.meta_data'))

    def create_user_data(self):
        self._user_data_file = self._create_tempfile_with_content(
            self.conf.get('vm.user_data'))

    def create_seed(self):
        if self._meta_data_file is None:
            self.create_meta_data()
        if self._user_data_file is None:
            self.create_user_data()
        # FIXME: seed_path leak (no big deal) -- vila 2013-02-01
        seed_path = os.path.join(self.conf.get('vm.images_dir'),
                                 self.conf.expand_options('{vm.name}.seed'))
        run_subprocess(
            # We create the seed in the ``vm.images_dir`` directory, so
            # ``sudo`` is required
            ['sudo',
             'genisoimage', '-output', seed_path, '-volid', 'cidata',
             '-joliet', '-rock', '-input-charset', 'default',
             '-graft-points',
             'user-data=%s' % (self._user_data_file.name,),
             'meta-data=%s' % (self._meta_data_file.name,),
             ])
        self._seed_path = seed_path

    def create_disk_image(self):
        raise NotImplementedError(self.create_disk_image)

    def _wait_for_install_with_seed(self):
        # The console is create by virt-install which requires sudo but creates
        # the file 0600 for libvirt-qemu. We give read access to all otherwise
        # 'tail -f' requires sudo and can't be killed anymore.
        run_subprocess(['sudo', 'chmod', '0644', self._console_path])
        # While `virt-install` is running, let's connect to the console
        console_cmd = ['tail', '-f', self._console_path]
        console = pipe_subprocess(console_cmd)
        lines = []
        while True:
            line = console.stdout.readline()
# FIXME: We need some way to activate this dynamically (conf var defaulting to
# env var OR cmdline parameter ? -- vila 2013-02-11
#            print "read: [%s]" % (line,) # so useful for debug...
            lines.append(line)
            if not line:
                # This should only happen when something goes wrong...
                raise CommandError(
                    console_cmd, console.returncode, '\n'.join(lines),
                    "EOF on console without expected final line")
            # FIXME: Use a regexp to be a bit less brittle (match on seed,
            # avoid the cloud-init version).

            # A typical install from seed on /dev/sdb looks like:

            # Cloud-init v. 0.7.1 finished at Tue, 05 Feb 2013 13:27:29 +0000. Datasource DataSourceNoCloudNet [seed=/dev/sdb][dsmode=net].  Up 7.86 seconds

            # so the regexp should be something like:
            # ^Cloud-init .* finished at .* Datasource DataSourceNoCloudNet (\[seed/dev/sdb\]) ? Up (.*) seconds

# precise: cloud-init boot finished at Tue, 12 Feb 2013 16:41:23 +0000. Up 2167.81 seconds
# quantal: Cloud-init v. 0.7 finished at Tue, 12 Feb 2013 16:50:04 +0000. Datasource DataSourceNoCloudNet [seed=/dev/sdb][dsmode=net].  Up 2236.07 seconds

            if line.startswith('Cloud-init v. 0.7.1 finished at'):
                break
            if ('Failed loading yaml blob' in line
                or 'Unhandled non-multipart userdata starting' in line):
                # Something was wrong in the configuration, report what we can
                # to help debug.
                err_lines = ['Suspicious line from cloud-init.\n',
                             'Check the configuration:\n']
                with open(self._meta_data_file.name) as f:
                    err_lines.append('meta-data content:\n')
                    err_lines.extend(f.readlines())
                with open(self._user_data_file.name) as f:
                    err_lines.append('user-data content:\n')
                    err_lines.extend(f.readlines())
                raise CommandError(console_cmd, console.returncode,
                                   '\n'.join(lines),
                                   ''.join(err_lines))
        console.terminate()

    def install(self):
        # Create a kvm, relying on cloud-init to customize the base image.
        #  
        # There are two processes involvded here:
        # - virt-install creates the vm and boots it.
        # - progress is monitored via the console to detect cloud-final.
        #  
        # Once cloud-init has finished, the vm can be powered off.

        # FIXME: If the install doesn't finish after $time, emit a warning and
        # terminate self.install_proc.
        # FIXME: If we can't connect to the console, emit a warning and
        # terminate console and self.install_proc.
        # FIXME: If we don't receive anything on the console after $time2, emit
        # a warning and terminate console and self.install_proc.
        # -- vila 2013-02-07
        if self._seed_path is None:
            self.create_seed()
        if self._disk_image_path is None:
            self.create_disk_image()
        self._console_path = os.path.join(
            self.conf.get('vm.images_dir'),
            '%s.console' % (self.conf.get('vm.name'),))
        run_subprocess(
            ['sudo', 'virt-install',
             # Without --noautoconsole, virt-install will relay the console,
             # that's not appropriate for our needs so we'll connect later
             # ourselves
             '--noautoconsole',
             # We define the console as a file so we can monitor the install
             # via 'tail -f'
             '--serial', 'file,path=%s' % (self._console_path,),
             '--network', self.conf.get('vm.network'),
             # Anticipate that we'll need a graphic card defined
             '--video', 'vga',
             '--name', self.conf.get('vm.name'),
             '--ram', self.conf.get('vm.ram_size'),
             '--vcpus', self.conf.get('vm.cpus'),
             '--disk', 'path=%s,format=qcow2' % (self._disk_image_path,),
             '--disk', 'path=%s' % (self._seed_path,),
             # We just boot, cloud-init will handle the installs we need
             '--boot', 'hd', '--hvm',
             ])
        self._wait_for_install_with_seed()

    def poweroff(self):
        return run_subprocess(
            ['sudo', 'virsh', 'destroy', self.conf.get('vm.name')])

    def undefine(self):
        return run_subprocess(
            ['sudo', 'virsh', 'undefine', self.conf.get('vm.name'),
             '--remove-all-storage'])


class KvmFromCloudImage(Kvm):

    def create_disk_image(self, src_name=None, dst_name=None):
        """Create a disk image from a cloud one."""
        if src_name is None:
            src_name = self.conf.get('vm.cloud_image_name')
        if dst_name is None:
            dst_name = self.conf.expand_options('{vm.name}.qcow2')
        cloud_image_path = os.path.join(
            self.conf.get('vm.download_cache'), src_name)
        disk_image_path = os.path.join(
            self.conf.get('vm.images_dir'), dst_name)
        run_subprocess(
            ['sudo', 'qemu-img', 'convert',
             '-O', 'qcow2', cloud_image_path, disk_image_path])
        run_subprocess(
            ['sudo', 'qemu-img', 'resize',
             disk_image_path, self.conf.get('vm.disk_size')])
        self._disk_image_path = disk_image_path


class KvmFromBacking(Kvm):

    def create_disk_image(self, src_name=None, dst_name=None):
        """Create a disk image backed by an existing one."""
        backing_image_path = os.path.join(
            self.conf.get('vm.images_dir'),
            self.conf.expand_options('{vm.backing}'))
        disk_image_path = os.path.join(
            self.conf.get('vm.images_dir'),
            self.conf.expand_options('{vm.name}.qcow2'))
        run_subprocess(
            ['sudo', 'qemu-img', 'create', '-f', 'qcow2',
             '-b', backing_image_path, disk_image_path])
        run_subprocess(
            ['sudo', 'qemu-img', 'resize',
             disk_image_path, self.conf.get('vm.disk_size')])
        self._disk_image_path = disk_image_path


class ArgParser(argparse.ArgumentParser):
    """A parser for the setup_vm script."""

    def __init__(self):
        description = 'Set up virtual machines from a configuration file.'
        super(ArgParser, self).__init__(
            prog='setup_vm.py', description=description)
        self.add_argument(
            'name', help='Virtual machine section in the configuration file.')
        self.add_argument('--download', '-d', action="store_true",
                          help='Force download of the required image.')
        self.add_argument('--install', '-i', action="store_true",
                          help='Install the virtual machine.')

    def parse_args(self, args=None, out=None, err=None):
        """Parse arguments, overridding stdout/stderr if provided.

        Overridding stdout/stderr is provided for tests.

        :params args: Defaults to sys.argv[1:].

        :param out: Defaults to sys.stdout.

        :param err: Defaults to sys.stderr.
        """
        out_orig = sys.stdout
        err_orig = sys.stderr
        try:
            if out is not None:
                sys.stdout = out
            if err is not None:
                sys.stderr = err
            return super(ArgParser, self).parse_args(args)
        finally:
            sys.stdout = out_orig
            sys.stderr = err_orig



arg_parser = ArgParser()

class Command(object):

    def __init__(self, vm):
        self.vm = vm


class Download(Command):

    def run(self):
        # FIXME: what needs to be downloaded should depend on the type of the
        # vm (possibly errors if there is nothing to download). -- vila
        # 2013-02-06
        self.vm.download_cloud_image(force=True)


class Install(Command):

    def run(self):
        vm_name = self.vm.conf.get('vm.name')
        state = vm_states().get(vm_name, None)
        if state == 'shut off':
            self.vm.undefine()
        elif state == 'running':
            raise SetupVmError('{name} is running', name=vm_name)
        # FIXME: The installation method may vary depending on the vm type.
        # -- vila 2013-02-06
        self.vm.install()
        self.vm.poweroff()


def build_commands(args=None, out=None, err=None):
    cmds = []
    if args is None:
        args = sys.argv[1:]

    ns = arg_parser.parse_args(args, out=out, err=err)

    conf = VmStack(ns.name)
    with_backing = conf.get('vm.backing')
    if with_backing is None:
        vm = KvmFromCloudImage(conf)
    else:
        vm = KvmFromBacking(conf)
    if ns.download:
        cmds.append(Download(vm))
    if ns.install:
        cmds.append(Install(vm))
    return cmds


def run(args=None):
    cmds = build_commands(args)
    for cmd in cmds:
        try:
            cmd.run()
        except SetupVmError, e:
            # Stop on first error
            print 'ERROR: %s' % e
            exit(-1)


if __name__ == "__main__":
    run()
