from cStringIO import StringIO
import os
import sys

import testtools

import tests
from bin import setup_vm


def requires_known_reference_image(test):
    # We need a pre-seeded download cache from the user running the tests
    # as downloading the cloud image is too long.
    user_conf = setup_vm.VmStack(None)
    download_cache = user_conf.get('vm.download_cache')
    if download_cache is None:
        test.skip('vm.download_cache is not set')
    # We use some known reference
    reference_cloud_image_name = 'raring-server-cloudimg-amd64-disk1.img'
    cloud_image_path = os.path.join(
        download_cache, reference_cloud_image_name)
    if not os.path.exists(cloud_image_path):
        test.skip('%s is not available' % (cloud_image_path,))
    return download_cache, reference_cloud_image_name


class TestCaseWithHome(testtools.TestCase):

    def setUp(self):
        super(TestCaseWithHome, self).setUp()
        tests.set_cwd_to_tmp(self)
        tests.isolate_env(self)
        self.home_dir = os.path.join(self.test_base_dir, 'home')
        os.mkdir(self.home_dir)
        os.environ['HOME'] = self.home_dir


class TestVmMatcher(TestCaseWithHome):

    def setUp(self):
        super(TestVmMatcher, self).setUp()
        self.store = setup_vm.VmStore('.', 'foo.conf')
        self.matcher = setup_vm.VmMatcher(self.store, 'test')

    def test_empty_section_always_matches(self):
        self.store._load_from_string('foo=bar')
        matching = list(self.matcher.get_sections())
        self.assertEqual(1, len(matching))

    def test_specific_before_generic(self):
        self.store._load_from_string('foo=bar\n[test]\nfoo=baz')
        matching = list(self.matcher.get_sections())
        self.assertEqual(2, len(matching))
        # First matching section is for test
        self.assertEqual(self.store, matching[0][0])
        base_section = matching[0][1]
        self.assertEqual('test', base_section.id)
        # Second matching section is the no-name one
        self.assertEqual(self.store, matching[0][0])
        no_name_section = matching[1][1]
        self.assertIs(None, no_name_section.id)


class TestVmStack(TestCaseWithHome):

    def setUp(self):
        super(TestVmStack, self).setUp()
        self.conf = setup_vm.VmStack('foo')
        self.conf.store._load_from_string('''
vm.release=raring
vm.cpu_model=amd64
''')

    def test_raring_iso_url(self):
        self.assertEqual('http://cdimage.ubuntu.com/daily-live/current/',
                         self.conf.get('vm.iso_url'))

    def test_raring_iso_name(self):
        self.assertEqual('raring-desktop-amd64.iso',
                         self.conf.get('vm.iso_name'))

    def test_raring_cloud_image_url(self):
        self.assertEqual('http://cloud-images.ubuntu.com/raring/current/',
                         self.conf.get('vm.cloud_image_url'))

    def test_raring_cloud_image_name(self):
        self.assertEqual('raring-server-cloudimg-amd64-disk1.img',
                         self.conf.get('vm.cloud_image_name'))

    def test_apt_proxy_not_set(self):
        self.assertEqual('', self.conf.expand_options('{vm.apt_proxy}'))

    def test_apt_proxy_set(self):
        proxy = 'apt_proxy: http://example.org:4321'
        self.conf.set('vm.apt_proxy', proxy)
        self.assertEqual(proxy, self.conf.expand_options('{vm.apt_proxy}'))



class TestDownload(TestCaseWithHome):

# FIXME: Needs parametrization against vm.{cloud_image_name|iso_name} and
# {download_iso|download_cloud_image} -- vila 2013-02-07

    def setUp(self):
        # Downloading real isos or images is too long for tests, instead, we
        # fake it by downloading a small but known to exist file: MD5SUMS
        super(TestDownload, self).setUp()
        download_cache = os.path.join(self.test_base_dir, 'downloads')
        os.mkdir(download_cache)
        self.conf = setup_vm.VmStack('foo')
        self.conf.store._load_from_string('''
vm.iso_name=MD5SUMS
vm.cloud_image_name=MD5SUMS
vm.release=raring
vm.cpu_model=amd64
vm.download_cache=%s
''' % (download_cache,))

    def test_download_iso(self):
        vm = setup_vm.Kvm(self.conf)
        self.assertTrue(vm.download_iso())
        # Trying to download again will find the file in the cache
        self.assertFalse(vm.download_iso())
        # Forcing the download even when the file is present
        self.assertTrue(vm.download_iso(force=True))

    def test_download_cloud_image(self):
        vm = setup_vm.Kvm(self.conf)
        self.assertTrue(vm.download_cloud_image())
        # Trying to download again will find the file in the cache
        self.assertFalse(vm.download_cloud_image())
        # Forcing the download even when the file is present
        self.assertTrue(vm.download_cloud_image(force=True))

    def test_download_unknown_iso_fail(self):
        self.conf.set('vm.iso_name', 'I-dont-exist')
        vm = setup_vm.Kvm(self.conf)
        self.assertRaises(setup_vm.CommandError, vm.download_iso)

    def test_download_unknown_cloud_image_fail(self):
        self.conf.set('vm.cloud_image_name', 'I-dont-exist')
        vm = setup_vm.Kvm(self.conf)
        self.assertRaises(setup_vm.CommandError, vm.download_cloud_image)

    def test_download_iso_with_unknown_cache_fail(self):
        dl_cache = os.path.join(self.test_base_dir, 'I-dont-exist')
        self.conf.set('vm.download_cache', dl_cache)
        vm = setup_vm.Kvm(self.conf)
        self.assertRaises(setup_vm.ConfigValueError, vm.download_iso)

    def test_download_cloud_image_with_unknown_cache_fail(self):
        dl_cache = os.path.join(self.test_base_dir, 'I-dont-exist')
        self.conf.set('vm.download_cache', dl_cache)
        vm = setup_vm.Kvm(self.conf)
        self.assertRaises(setup_vm.ConfigValueError, vm.download_cloud_image)


class TestSeed(TestCaseWithHome):

    def setUp(self):
        super(TestSeed, self).setUp()
        self.conf = setup_vm.VmStack('foo')
        self.vm = setup_vm.Kvm(self.conf)
        images_dir = os.path.join(self.test_base_dir, 'images')
        os.mkdir(images_dir)
        self.conf.store._load_from_string('''
vm.name=foo
vm.ssh_pub_key=bar
vm.release=raring
vm.images_dir=%s
''' % (images_dir,))

    def assertPathCleanedUp(self, tmp_path):
        self.assertTrue(os.path.exists(tmp_path))
        # The temp file is deleted when the vm object is
        del self.vm
        self.assertFalse(os.path.exists(tmp_path))

    def test_create_meta_data(self):
        self.vm.create_meta_data()
        self.assertPathCleanedUp(self.vm._meta_data_file.name)

    def test_create_user_data(self):
        self.vm.create_user_data()
        self.assertPathCleanedUp(self.vm._user_data_file.name)

    def test_create_seed(self):
        self.assertTrue(self.vm._seed_path is None)
        self.vm.create_seed()
        self.assertFalse(self.vm._seed_path is None)
        self.assertTrue(os.path.exists(self.vm._seed_path))


class TestImageFromCloud(TestCaseWithHome):

    def setUp(self):
        super(TestImageFromCloud, self).setUp()
        self.conf = setup_vm.VmStack('foo')
        self.vm = setup_vm.KvmFromCloudImage(self.conf)
        images_dir = os.path.join(self.test_base_dir, 'images')
        os.mkdir(images_dir)
        download_cache_dir = os.path.join(self.test_base_dir, 'download')
        os.mkdir(download_cache_dir)
        self.conf.store._load_from_string('''
vm.name=foo
vm.ssh_pub_key=bar
vm.release=raring
vm.images_dir=%s
vm.download_cache=%s
vm.cloud_image_name=fake.img
vm.disk_size=1M
''' % (images_dir, download_cache_dir))

    def test_create_disk_image(self):
        cloud_image_path = os.path.join(self.conf.get('vm.download_cache'),
                                        self.conf.get('vm.cloud_image_name'))
        # We need a fake cloud image that can be converted
        setup_vm.run_subprocess(
            ['sudo', 'qemu-img', 'create',
             cloud_image_path, self.conf.get('vm.disk_size')])
        self.assertTrue(self.vm._disk_image_path is None)
        self.vm.create_disk_image()
        self.assertFalse(self.vm._disk_image_path is None)
        self.assertTrue(os.path.exists(self.vm._disk_image_path))


class TestImageWithBacking(TestCaseWithHome):

    def setUp(self):
        (download_cache_dir,
         reference_cloud_image_name) = requires_known_reference_image(self)
        super(TestImageWithBacking, self).setUp()
        # We'll share the images_dir between vms
        images_dir = os.path.join(self.test_base_dir, 'images')
        os.mkdir(images_dir)
        # Create a shared config
        conf = setup_vm.VmStack(None)
        conf.store._load_from_string('''
vm.ssh_pub_key=bar
vm.release=raring
vm.images_dir=%s
vm.download_cache=%s
vm.disk_size=2G
[selftest-from-cloud]
vm.name=selftest-from-cloud
vm.cloud_image_name=%s
[selftest-backing]
vm.name=selftest-backing
vm.backing=selftest-from-cloud.qcow2
''' % (images_dir, download_cache_dir, reference_cloud_image_name))
        conf.store.save()
        # To bypass creating a real vm, we start from the cloud image that is a
        # real and bootable one, so we just convert it. That also makes it
        # available in vm.images_dir
        temp_vm = setup_vm.KvmFromCloudImage(
            setup_vm.VmStack('selftest-from-cloud'))
        temp_vm.create_disk_image()

    def test_create_image_with_backing(self):
        vm = setup_vm.KvmFromBacking(setup_vm.VmStack('selftest-backing'))
        self.assertTrue(vm._disk_image_path is None)
        vm.create_disk_image()
        self.assertFalse(vm._disk_image_path is None)
        self.assertTrue(os.path.exists(vm._disk_image_path))


class TestVmStates(testtools.TestCase):

    def assertStates(self, expected, lines):
        self.assertEqual(expected, setup_vm.vm_states(lines))

    def test_empty(self):
        self.assertStates({},[])

    def test_garbage(self):
        self.assertRaises(ValueError, self.assertStates, None, [''])

    def test_known_states(self):
        # From a real life sample
        self.assertStates({'foo': 'shut off', 'bar': 'running'},
                          ['-     foo                            shut off',
                           '19    bar                            running'])


class TestInstallWithSeed(TestCaseWithHome):

    def setUp(self):
        (download_cache,
         reference_cloud_image_name) = requires_known_reference_image(self)
        super(TestInstallWithSeed, self).setUp()
        # We need to allow other users to read this dir
        os.chmod(self.test_base_dir, 0755)
        # We also need to sudo rm it as root created some files there
        self.addCleanup(
            setup_vm.run_subprocess,
            ['sudo', 'rm', '-fr',
             os.path.join(self.test_base_dir, 'home', '.virtinst')])
        self.conf = setup_vm.VmStack('selftest-seed')
        self.vm = setup_vm.KvmFromCloudImage(self.conf)
        images_dir = os.path.join(self.test_base_dir, 'images')
        os.mkdir(images_dir, 0755)
        # Setup user-data with a config that leads to a very short install time
        user_data = '''\
#cloud-config
locale: en_US.UTF-8
password: {vm.password}
chpasswd: { expire: False }
apt_update: false
apt_upgrade: false
'''
        self.conf.store._load_from_string('''
vm.name=selftest-seed
vm.cpus=2,
vm.ssh_pub_key=bar
vm.release=raring
vm.images_dir=%s
vm.download_cache=%s
vm.cloud_image_name=%s
vm.disk_size=8G
vm.user_data="""%s"""
''' % (images_dir, download_cache, reference_cloud_image_name, user_data))

    def assertVmState(self, expected):
        states = setup_vm.vm_states()
        self.assertEqual(expected, states[self.vm.conf.get('vm.name')])

    def test_install_with_seed(self):
        self.addCleanup(self.vm.undefine)
        self.vm.install()
        self.assertVmState('running')
        self.vm.poweroff()
        self.assertVmState('shut off')


class TestInstallWithBacking(TestCaseWithHome):

    def setUp(self):
        (download_cache_dir,
         reference_cloud_image_name) = requires_known_reference_image(self)
        super(TestInstallWithBacking, self).setUp()
        # We need to allow other users to read this dir
        os.chmod(self.test_base_dir, 0755)
        # We also need to sudo rm it as root created some files there
        self.addCleanup(
            setup_vm.run_subprocess,
            ['sudo', 'rm', '-fr',
             os.path.join(self.test_base_dir, 'home', '.virtinst')])
        self.conf = setup_vm.VmStack('selftest-backing')
        self.vm = setup_vm.KvmFromBacking(self.conf)
        # We'll share the images_dir between vms
        images_dir = os.path.join(self.test_base_dir, 'images')
        os.mkdir(images_dir, 0755)
        # Setup user-data with a config that leads to a very short install time
        user_data = '''\
#cloud-config
locale: en_US.UTF-8
password: {vm.password}
chpasswd: { expire: False }
apt_update: false
apt_upgrade: false
'''
        # Create a shared config
        conf = setup_vm.VmStack(None)
        conf.store._load_from_string('''
vm.ssh_pub_key=bar
vm.release=raring
vm.images_dir=%s
vm.download_cache=%s
vm.disk_size=2G
vm.user_data="""%s"""
[selftest-from-cloud]
vm.name=selftest-from-cloud
vm.cloud_image_name=%s
[selftest-backing]
vm.name=selftest-backing
vm.backing=selftest-from-cloud.qcow2
''' % (images_dir, download_cache_dir, user_data, reference_cloud_image_name))
        conf.store.save()
        # Fake a previous install by just re-using the reference cloud image
        temp_vm = setup_vm.KvmFromCloudImage(
            setup_vm.VmStack('selftest-from-cloud'))
        temp_vm.create_disk_image()

    def assertVmState(self, vm, expected):
        states = setup_vm.vm_states()
        self.assertEqual(expected, states[vm.conf.get('vm.name')])

    def test_install_with_backing(self):
        vm = setup_vm.KvmFromBacking(setup_vm.VmStack('selftest-backing'))
        self.addCleanup(vm.undefine)
        vm.install()
        self.assertVmState(vm, 'running')
        vm.poweroff()
        self.assertVmState(vm, 'shut off')


class TestOptionParsing(testtools.TestCase):

    def setUp(self):
        super(TestOptionParsing, self).setUp()
        self.out = StringIO()
        self.err = StringIO()

    def parse_args(self, args):
        return setup_vm.arg_parser.parse_args(args, self.out, self.err)

    def test_nothing(self):
        self.assertRaises(SystemExit, self.parse_args, [])

    def test_install(self):
        ns = self.parse_args(['foo', '--install'])
        self.assertEquals('foo', ns.name)
        self.assertTrue(ns.install)
        self.assertFalse(ns.download)

    def test_download(self):
        ns = self.parse_args(['foo', '--download'])
        self.assertEquals('foo', ns.name)
        self.assertFalse(ns.install)
        self.assertTrue(ns.download)

class TestBuildCommands(testtools.TestCase):

    def setUp(self):
        super(TestBuildCommands, self).setUp()
        self.out = StringIO()
        self.err = StringIO()

    def build_commands(self, args):
        return setup_vm.build_commands(args, self.out, self.err)

    def test_install(self):
        cmds = self.build_commands(['foo', '--install'])
        self.assertEqual(1, len(cmds))
        self.assertTrue(isinstance(cmds[0], setup_vm.Install))

    def test_download(self):
        cmds = self.build_commands(['foo', '--download'])
        self.assertEqual(1, len(cmds))
        self.assertTrue(isinstance(cmds[0], setup_vm.Download))

    def test_download_and_install(self):
        cmds = self.build_commands(['foo', '--install', '--download'])
        self.assertEqual(2, len(cmds))
        # Download comes first
        self.assertTrue(isinstance(cmds[0], setup_vm.Download))
        self.assertTrue(isinstance(cmds[1], setup_vm.Install))


# FIXME: This needs to be parametrized for KvmFromCloudImage and
# KvmFromBacking. Since we don't define vm.backing below, we're only testing
# KvmFromCloudImage for now. -- vila 2013-02-13
class TestInstall(TestCaseWithHome):

    def setUp(self):
        super(TestInstall, self).setUp()
        self.conf = setup_vm.VmStack('I-dont-exist')
        self.conf.store._load_from_string('''
vm.name=I-dont-exist
vm.user_data=''
vm.release=raring
vm.cpu_model=amd64
''')
        self.states = []

        def vm_states(source=None):
            return self.states
        self.patch(setup_vm, 'vm_states', vm_states)
        self.vm = None

    def install(self):
        class FakeKvm(setup_vm.Kvm):

            def __init__(self, conf):
                super(FakeKvm, self).__init__(conf)
                self.undefine_called = False
                self.install_called = False

            # Make sure we avoid dangerous or costly calls
            def poweroff(self):
                pass

            def undefine(self):
                self.undefine_called = True

            def install(self):
                self.install_called = True


        self.vm = FakeKvm(self.conf)
        cmd = setup_vm.Install(self.vm)
        cmd.run()

    def test_install_while_running(self):
        self.conf.set('vm.name', 'foo')
        self.states = {'foo': 'running'}
        self.assertRaises(setup_vm.SetupVmError, self.install)
        self.assertFalse(self.vm.install_called)
        self.assertFalse(self.vm.undefine_called)

    def test_install_unknown(self):
        self.states = {}
        self.install()
        self.assertTrue(self.vm.install_called)
        self.assertFalse(self.vm.undefine_called)

    def test_install_shutoff(self):
        self.conf.set('vm.name', 'foo')
        self.states = {'foo': 'shut off'}
        self.install()
        self.assertTrue(self.vm.install_called)
        self.assertTrue(self.vm.undefine_called)
